package dev.shadowsoffire.apotheosis.spawn.modifiers;

import net.minecraft.util.Mth;
import dev.shadowsoffire.apotheosis.spawn.spawner.ApothSpawnerBlock;
import net.minecraft.network.chat.CommonComponents;
import net.minecraft.ChatFormatting;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import com.mojang.datafixers.kinds.Applicative;
import com.mojang.datafixers.kinds.App;
import com.mojang.serialization.Codec;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.LinkedHashMap;
import java.util.Iterator;
import net.minecraft.network.chat.Component;
import java.util.function.Consumer;
import dev.shadowsoffire.apotheosis.spawn.spawner.ApothSpawnerTile;
import java.util.Map;

public class SpawnerStats
{
    public static final Map<String, SpawnerStat<?>> REGISTRY;
    public static final SpawnerStat<Short> MIN_DELAY;
    public static final SpawnerStat<Short> MAX_DELAY;
    public static final SpawnerStat<Short> SPAWN_COUNT;
    public static final SpawnerStat<Short> MAX_NEARBY_ENTITIES;
    public static final SpawnerStat<Short> REQ_PLAYER_RANGE;
    public static final SpawnerStat<Short> SPAWN_RANGE;
    public static final SpawnerStat<Boolean> IGNORE_PLAYERS;
    public static final SpawnerStat<Boolean> IGNORE_CONDITIONS;
    public static final SpawnerStat<Boolean> REDSTONE_CONTROL;
    public static final SpawnerStat<Boolean> IGNORE_LIGHT;
    public static final SpawnerStat<Boolean> NO_AI;
    public static final SpawnerStat<Boolean> SILENT;
    public static final SpawnerStat<Boolean> BABY;
    
    public static void generateTooltip(final ApothSpawnerTile tile, final Consumer<Component> list) {
        for (final SpawnerStat<?> stat : SpawnerStats.REGISTRY.values()) {
            final Component comp = stat.getTooltip(tile);
            if (!comp.getString().isEmpty()) {
                list.accept(comp);
            }
        }
    }
    
    private static <T extends SpawnerStat<?>> T register(final T t) {
        SpawnerStats.REGISTRY.put(t.getId(), t);
        return t;
    }
    
    static {
        REGISTRY = new LinkedHashMap<String, SpawnerStat<?>>();
        MIN_DELAY = register(new ShortStat("min_delay", s -> s.f_59788_.f_45447_, (s, v) -> s.f_59788_.f_45447_ = v));
        MAX_DELAY = register(new ShortStat("max_delay", s -> s.f_59788_.f_45448_, (s, v) -> s.f_59788_.f_45448_ = v));
        SPAWN_COUNT = register(new ShortStat("spawn_count", s -> s.f_59788_.f_45449_, (s, v) -> s.f_59788_.f_45449_ = v));
        MAX_NEARBY_ENTITIES = register(new ShortStat("max_nearby_entities", s -> s.f_59788_.f_45451_, (s, v) -> s.f_59788_.f_45451_ = v));
        REQ_PLAYER_RANGE = register(new ShortStat("req_player_range", s -> s.f_59788_.f_45452_, (s, v) -> s.f_59788_.f_45452_ = v));
        SPAWN_RANGE = register(new ShortStat("spawn_range", s -> s.f_59788_.f_45453_, (s, v) -> s.f_59788_.f_45453_ = v));
        IGNORE_PLAYERS = register(new BoolStat("ignore_players", s -> s.ignoresPlayers, (s, v) -> s.ignoresPlayers = v));
        IGNORE_CONDITIONS = register(new BoolStat("ignore_conditions", s -> s.ignoresConditions, (s, v) -> s.ignoresConditions = v));
        REDSTONE_CONTROL = register(new BoolStat("redstone_control", s -> s.redstoneControl, (s, v) -> s.redstoneControl = v));
        IGNORE_LIGHT = register(new BoolStat("ignore_light", s -> s.ignoresLight, (s, v) -> s.ignoresLight = v));
        NO_AI = register(new BoolStat("no_ai", s -> s.hasNoAI, (s, v) -> s.hasNoAI = v));
        SILENT = register(new BoolStat("silent", s -> s.silent, (s, v) -> s.silent = v));
        BABY = register(new BoolStat("baby", s -> s.baby, (s, v) -> s.baby = v));
    }
    
    private abstract static class Base<T> implements SpawnerStat<T>
    {
        protected final String id;
        protected final Function<ApothSpawnerTile, T> getter;
        protected final BiConsumer<ApothSpawnerTile, T> setter;
        
        private Base(final String id, final Function<ApothSpawnerTile, T> getter, final BiConsumer<ApothSpawnerTile, T> setter) {
            this.id = id;
            this.getter = getter;
            this.setter = setter;
        }
        
        @Override
        public String getId() {
            return this.id;
        }
        
        @Override
        public T getValue(final ApothSpawnerTile spawner) {
            return this.getter.apply(spawner);
        }
    }
    
    private static class BoolStat extends Base<Boolean>
    {
        private final Codec<StatModifier<Boolean>> modifierCodec;
        
        private BoolStat(final String id, final Function<ApothSpawnerTile, Boolean> getter, final BiConsumer<ApothSpawnerTile, Boolean> setter) {
            super(id, getter, setter);
            this.modifierCodec = (Codec<StatModifier<Boolean>>)RecordCodecBuilder.create(inst -> inst.group((App)Codec.BOOL.fieldOf("value").forGetter((Function)StatModifier::value)).apply((Applicative)inst, value -> new StatModifier((SpawnerStat<Object>)this, value, false, true)));
        }
        
        @Override
        public Codec<StatModifier<Boolean>> getModifierCodec() {
            return this.modifierCodec;
        }
        
        @Override
        public Component getTooltip(final ApothSpawnerTile spawner) {
            return (Component)(((boolean)this.getValue(spawner)) ? this.name().m_130940_(ChatFormatting.DARK_GREEN) : CommonComponents.f_237098_);
        }
        
        @Override
        public boolean apply(final Boolean value, final Boolean min, final Boolean max, final ApothSpawnerTile spawner) {
            final boolean old = (boolean)this.getter.apply(spawner);
            this.setter.accept(spawner, (T)value);
            return old != (boolean)this.getter.apply(spawner);
        }
    }
    
    private static class ShortStat extends Base<Short>
    {
        public static final Codec<Short> BOUNDS_CODEC;
        private final Codec<StatModifier<Short>> modifierCodec;
        
        private ShortStat(final String id, final Function<ApothSpawnerTile, Integer> getter, final BiConsumer<ApothSpawnerTile, Short> setter) {
            super(id, tile -> getter.apply(tile).shortValue(), setter);
            this.modifierCodec = (Codec<StatModifier<Short>>)RecordCodecBuilder.create(inst -> inst.group((App)Codec.SHORT.fieldOf("value").forGetter((Function)StatModifier::value), (App)ShortStat.BOUNDS_CODEC.fieldOf("min").forGetter((Function)StatModifier::min), (App)ShortStat.BOUNDS_CODEC.fieldOf("max").forGetter((Function)StatModifier::max)).apply((Applicative)inst, (value, min, max) -> new StatModifier((SpawnerStat<T>)this, (T)value, (T)(short)((min == -1) ? 0 : ((short)min)), (T)(short)((max == -1) ? 32767 : ((short)max)))));
        }
        
        @Override
        public Codec<StatModifier<Short>> getModifierCodec() {
            return this.modifierCodec;
        }
        
        @Override
        public Component getTooltip(final ApothSpawnerTile spawner) {
            return ApothSpawnerBlock.concat(new Object[] { this.name(), ((Base<Object>)this).getValue(spawner) });
        }
        
        @Override
        public boolean apply(final Short value, final Short min, final Short max, final ApothSpawnerTile spawner) {
            final int old = (short)this.getter.apply(spawner);
            this.setter.accept(spawner, (T)(short)Mth.m_14045_(old + value, (int)min, (int)max));
            return old != (short)this.getter.apply(spawner);
        }
        
        static {
            BOUNDS_CODEC = Codec.intRange(-1, 32767).xmap((Function)Integer::shortValue, (Function)Short::intValue);
        }
    }
}
